"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology

This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import json
import scipy.sparse
from servalcat.utils import logger
from servalcat.xtal import sigmaa
from servalcat import utils
from servalcat import ext
from servalcat.xtal.twin import find_twin_domains_from_data, estimate_twin_fractions_from_model
b_to_u = utils.model.b_to_u
u_to_b = utils.model.u_to_b
integr = sigmaa.integr

class LL_Xtal:
    def __init__(self, hkldata, free, st, monlib, source="xray", mott_bethe=True,
                 use_solvent=False, use_in_est="all", use_in_target="all", twin=False, is_int=None):
        assert source in ("electron", "xray", "neutron")
        self.source = source
        self.mott_bethe = False if source != "electron" else mott_bethe
        self.hkldata = hkldata
        self.free = free
        self.st = st
        self.monlib = monlib
        self.d_min_max = hkldata.d_min_max()
        self.fc_labs = ["FC0"]
        self.use_solvent = use_solvent
        if use_solvent:
            self.fc_labs.append("FCbulk")
            self.hkldata.df["FCbulk"] = 0j
        self.D_labs = ["D{}".format(i) for i in range(len(self.fc_labs))]
        self.k_overall = numpy.ones(len(self.hkldata.df.index))
        self.b_aniso = gemmi.SMat33d(0,0,0,0,0,0)
        self.hkldata.df["k_aniso"] = 1.
        self.use_in_est = use_in_est
        self.use_in_target = use_in_target
        self.ll = None
        self.scaling = sigmaa.LsqScale()
        if twin:
            self.twin_data, _ = find_twin_domains_from_data(self.hkldata)
        else:
            self.twin_data = None
        if self.twin_data:
            self.twin_data.setup_f_calc(len(self.fc_labs))
        if is_int is None:
            self.is_int = "I" in self.hkldata.df
        else:
            self.is_int = is_int
        logger.writeln("will use {} reflections for parameter estimation".format(self.use_in_est))
        logger.writeln("will use {} reflections for refinement".format(self.use_in_target))

    def refine_id(self):
        return {"xray": "X-RAY", "electron": "ELECTRON", "neutron": "NEUTRON"}.get(self.source, "") + " DIFFRACTION"

    def update_ml_params(self):
        self.b_aniso = sigmaa.determine_ml_params(self.hkldata, self.is_int, self.fc_labs, self.D_labs, self.b_aniso,
                                                  use=self.use_in_est,
                                                  twin_data=self.twin_data)#D_trans="splus", S_trans="splus")
        self.hkldata.df["k_aniso"] = self.hkldata.debye_waller_factors(b_cart=self.b_aniso)
        #determine_mlf_params_from_cc(self.hkldata, self.fc_labs, self.D_labs)
    def update_fc(self):
        # modify st before fc calculation
        b_resid = sigmaa.subtract_common_aniso_from_model([self.st])
        self.b_aniso += gemmi.SMat33d(*b_resid.elements_pdb()) # needed for target calculation
        d_min = max(self.twin_data.s2_array)**(-0.5) if self.twin_data else self.d_min_max[0]
        sigmaa.update_fc(st_list=[self.st], fc_labs=self.fc_labs,
                         d_min=d_min, monlib=self.monlib,
                         source=self.source, mott_bethe=self.mott_bethe,
                         hkldata=self.hkldata, twin_data=self.twin_data)

    def prepare_target(self):
        if self.twin_data:
            if self.use_in_target == "all":
                idxes = numpy.concatenate([sel[i] for i_bin, _ in self.hkldata.binned("ml")
                                           for sel in self.hkldata.centric_and_selections["ml"][i_bin] for i in (1,2)])
            else:
                i = 1 if self.use_in_target == "work" else 2
                idxes = numpy.concatenate([sel[i] for i_bin, _ in self.hkldata.binned("ml")
                                           for sel in self.hkldata.centric_and_selections["ml"][i_bin]])
            mask = numpy.empty(len(self.hkldata.df.index)) * numpy.nan
            mask[idxes] = 1 / self.hkldata.debye_waller_factors(b_cart=self.b_aniso)[idxes]**2
            self.twin_data.est_f_true(self.hkldata.df.I.to_numpy() * mask,
                                      self.hkldata.df.SIGI.to_numpy() * mask)
            self.k_ani2_inv_masked = mask
        
    def _overall_scale(self, min_b=0.1):
        miller_array = self.twin_data.asu if self.twin_data else self.hkldata.miller_array()
        d_min = max(self.twin_data.s2_array)**(-0.5) if self.twin_data else self.d_min_max[0]
        if self.use_solvent:
            Fmask = sigmaa.calc_Fmask(self.st, d_min, miller_array)
            if self.twin_data:
                fc_sum = self.twin_data.f_calc[:,:-1].sum(axis=1)
            else:
                fc_sum = self.hkldata.df[self.fc_labs[:-1]].sum(axis=1).to_numpy()
            fc_list = [fc_sum, Fmask]
        else:
            if self.twin_data:
                fc_list = [self.twin_data.f_calc.sum(axis=1)]
            else:
                fc_list = [self.hkldata.df[self.fc_labs].sum(axis=1).to_numpy()]
        self.scaling.set_data(self.hkldata, fc_list, self.is_int, sigma_cutoff=0, twin_data=self.twin_data)
        self.scaling.scale()
        self.b_aniso = self.scaling.b_aniso
        b = self.scaling.b_iso
        min_b_iso = self.st[0].calculate_b_aniso_range()[0] # actually min of aniso too
        tmp = min_b_iso + b
        if tmp < min_b: # perhaps better only adjust b_iso that went too small, but we need to recalculate Fc
            logger.writeln(" Adjusting overall B to avoid too small value")
            b += min_b - tmp
        logger.writeln(" Applying overall B to model: {:.2f}".format(b))
        utils.model.shift_b(self.st[0], b)
        k_iso = self.hkldata.debye_waller_factors(b_iso=b)
        self.hkldata.df["k_aniso"] = self.hkldata.debye_waller_factors(b_cart=self.b_aniso)
        if self.use_solvent:
            if self.twin_data:
                s2 = numpy.asarray(self.twin_data.s2_array)
            else:
                s2 = 1. / self.hkldata.d_spacings().to_numpy()**2
            Fbulk = Fmask * self.scaling.get_solvent_scale(self.scaling.k_sol, self.scaling.b_sol, s2)
            if self.twin_data:
                self.twin_data.f_calc[:,-1] = Fbulk
            else:
                self.hkldata.df[self.fc_labs[-1]] = Fbulk
        if "I" in self.hkldata.df:
            o_labs = self.hkldata.df.columns.intersection(["I", "SIGI",
                                                           "I(+)","SIGI(+)", "I(-)", "SIGI(-)"])
            self.hkldata.df[o_labs] /= self.scaling.k_overall**2
        if "FP" in self.hkldata.df:
            o_labs = self.hkldata.df.columns.intersection(["FP", "SIGFP",
                                                           "F(+)","SIGF(+)", "F(-)", "SIGF(-)"])
            self.hkldata.df[o_labs] /= self.scaling.k_overall

        if self.twin_data:
            self.twin_data.f_calc[:] *= self.twin_data.debye_waller_factors(b_iso=b)[:,None]
        else:
            for lab in self.fc_labs: self.hkldata.df[lab] *= k_iso
            self.hkldata.df["FC"] = self.hkldata.df[self.fc_labs].sum(axis=1)

        # for next cycle
        self.scaling.k_overall = 1.
        self.scaling.b_iso = 0.
    # _overall_scale()
    
    def overall_scale(self, min_b=0.1):
        self._overall_scale(min_b)
        if self.twin_data:
            estimate_twin_fractions_from_model(self.twin_data, self.hkldata)
            self._overall_scale(min_b)
    # overall_scale()

    def calc_target(self): # -LL target for MLF or MLI
        ret = 0
        if self.twin_data:
            Io = self.hkldata.df.I.to_numpy() * self.k_ani2_inv_masked
            sigIo = self.hkldata.df.SIGI.to_numpy() * self.k_ani2_inv_masked
            self.twin_data.est_f_true(Io, sigIo)
            ret = self.twin_data.ll(Io, sigIo)
        else:
            k_aniso = self.hkldata.debye_waller_factors(b_cart=self.b_aniso)
            f = sigmaa.mli if self.is_int else sigmaa.mlf
            for i_bin, _ in self.hkldata.binned("ml"):
                if self.use_in_target == "all":
                    idxes = numpy.concatenate([sel[i] for sel in self.hkldata.centric_and_selections["ml"][i_bin] for i in (1,2)])
                else:
                    i = 1 if self.use_in_target == "work" else 2
                    idxes = numpy.concatenate([sel[i] for sel in self.hkldata.centric_and_selections["ml"][i_bin]])
                ret += f(self.hkldata.df,
                         self.fc_labs,
                         numpy.vstack([self.hkldata.df[lab].to_numpy()[idxes] for lab in self.D_labs]).T,
                         self.hkldata.df.S.to_numpy()[idxes],
                         k_aniso,
                         idxes)
        return ret * 2 # friedel mates
    # calc_target()

    def calc_stats(self, bin_stats=False):
        stats, overall = sigmaa.calc_r_and_cc(self.hkldata, self.twin_data)
        ret = {"summary": overall}
        ret["summary"]["-LL"] = self.calc_target()
        if self.twin_data:
            ret["twin_alpha"] = {op.as_hkl().triplet(): a
                                 # ops does not include identity
                                 for op, a in zip([gemmi.Op()]+self.twin_data.ops, self.twin_data.alphas)}
        if bin_stats:
            ret["bin_stats"] = stats
            ret["ml"] = self.hkldata.binned_df["ml"].copy()
        for lab in "R", "CC":
            logger.writeln(" ".join("{} = {:.4f}".format(x, overall[x]) for x in overall if x.startswith(lab)))
        if self.is_int:
            logger.writeln("R1 is calculated for reflections with I/sigma>2.")
        return ret

    def calc_grad(self, refine_params, specs=None):
        blur = utils.model.determine_blur_for_dencalc(self.st, self.d_min_max[0] / 3) # TODO need more work
        logger.writeln("blur for deriv= {:.2f}".format(blur))
        if self.twin_data:
            Io = self.hkldata.df.I.to_numpy() * self.k_ani2_inv_masked
            sigIo = self.hkldata.df.SIGI.to_numpy() * self.k_ani2_inv_masked
            dll_dab, d2ll_dab2 = self.twin_data.ll_der_fc0(Io, sigIo)
            dll_dab *= self.twin_data.debye_waller_factors(b_iso=-blur)
        else:
            dll_dab = numpy.zeros(len(self.hkldata.df.FC), dtype=numpy.complex128)
            d2ll_dab2 = numpy.empty(len(self.hkldata.df.index))
            d2ll_dab2[:] = numpy.nan
            k_ani = self.hkldata.debye_waller_factors(b_cart=self.b_aniso)
            for i_bin, _ in self.hkldata.binned("ml"):
                for c, work, test in self.hkldata.centric_and_selections["ml"][i_bin]:
                    if self.use_in_target == "all":
                        cidxes = numpy.concatenate([work, test])
                    else:
                        cidxes = work if self.use_in_target == "work" else test
                    epsilon = self.hkldata.df.epsilon.to_numpy()[cidxes]
                    Fcs = numpy.vstack([self.hkldata.df[lab].to_numpy()[cidxes] for lab in self.fc_labs]).T
                    Ds = numpy.vstack([self.hkldata.df[lab].to_numpy()[cidxes] for lab in self.D_labs]).T
                    S = self.hkldata.df["S"].to_numpy()[cidxes]
                    Fc = (Ds * Fcs).sum(axis=1)
                    Fc_abs = numpy.abs(Fc)
                    expip = numpy.exp(1j * numpy.angle(Fc))
                    if self.is_int:
                        Io = self.hkldata.df.I.to_numpy()
                        sigIo = self.hkldata.df.SIGI.to_numpy()
                        to = Io[cidxes] / sigIo[cidxes] - sigIo[cidxes] / (c+1) / k_ani[cidxes]**2 / S / epsilon
                        tf = k_ani[cidxes] * Fc_abs / numpy.sqrt(sigIo[cidxes])
                        sig1 = k_ani[cidxes]**2 * epsilon * S / sigIo[cidxes]
                        k_num = numpy.repeat(0.5 if c == 0 else 0., to.size) # acentric:0.5, centric: 0.
                        r = ext.integ_J_ratio(k_num, k_num - 0.5, True, to, tf, sig1, numpy.repeat(c+1, to.size),
                                              integr.exp2_threshold, integr.h, integr.N, integr.ewmax)
                        r *= numpy.sqrt(sigIo[cidxes]) / k_ani[cidxes]
                        g = (2-c) * (Fc_abs - r) / epsilon / S  * Ds[:,0]
                        dll_dab[cidxes] = g * expip
                        #d2ll_dab2[cidxes] = (2-c)**2 / S / epsilon * Ds[0]**2 # approximation
                        #d2ll_dab2[cidxes] = ((2-c) / S / epsilon + ((2-c) * r / k_ani[cidxes] / epsilon / S)**2) * Ds[0]**2
                        d2ll_dab2[cidxes] =  g**2
                    else:
                        Fo = self.hkldata.df.FP.to_numpy()[cidxes] / k_ani[cidxes]
                        SigFo = self.hkldata.df.SIGFP.to_numpy()[cidxes] / k_ani[cidxes]
                        if c == 0: # acentric
                            Sigma = 2 * SigFo**2 + epsilon * S
                            X = 2 * Fo * Fc_abs / Sigma
                            m = gemmi.bessel_i1_over_i0(X)
                            g = 2 * (Fc_abs - m * Fo) / Sigma * Ds[:,0]  # XXX assuming 0 is atomic structure
                            dll_dab[cidxes] = g * expip
                            d2ll_dab2[cidxes] = (2 / Sigma - (1 - m / X - m**2) * (2 * Fo / Sigma)**2) * Ds[:,0]**2
                        else:
                            Sigma = SigFo**2 + epsilon * S
                            X = Fo * Fc_abs / Sigma
                            #X = X.astype(numpy.float64)
                            m = numpy.tanh(X)
                            g = (Fc_abs - m * Fo) / Sigma * Ds[:,0]
                            dll_dab[cidxes] = g * expip
                            d2ll_dab2[cidxes] = (1. / Sigma - (Fo / Sigma)**2 * (1. - m**2)) * Ds[:,0]**2
            dll_dab *= self.hkldata.debye_waller_factors(b_iso=-blur)

        if self.mott_bethe:
            d2 = numpy.reciprocal(self.twin_data.s2_array) if self.twin_data else self.hkldata.d_spacings()**2
            dll_dab *= d2 * gemmi.mott_bethe_const()
            d2ll_dab2 *= gemmi.mott_bethe_const()**2

        if not self.twin_data:
            dll_dab *= self.hkldata.df.llweight
            d2ll_dab2 *= self.hkldata.df.llweight
            
        # we need V**2/n for gradient.
        if self.twin_data:
            dll_dab_den = utils.hkl.fft_map(self.hkldata.cell, self.hkldata.sg, self.twin_data.asu, data=dll_dab)
        else:
            dll_dab_den = self.hkldata.fft_map(data=dll_dab)
        dll_dab_den.array[:] *= self.hkldata.cell.volume**2 / dll_dab_den.point_count
        #asu = dll_dab_den.masked_asu()
        #dll_dab_den.array[:] *= 1 - asu.mask_array # 0 to use
        
        self.ll = ext.LL(self.st, refine_params, self.mott_bethe)
        self.ll.set_ncs([x.tr for x in self.st.ncs if not x.given])
        if self.source == "neutron":
            self.ll.calc_grad_n92(dll_dab_den, blur)
        else:
            self.ll.calc_grad_it92(dll_dab_den, blur)

        # second derivative
        s_array = numpy.sqrt(self.twin_data.s2_array) if self.twin_data else 1./self.hkldata.d_spacings().to_numpy()
        if self.source == "neutron":
            self.ll.make_fisher_table_diag_direct_n92(s_array, d2ll_dab2)
            self.ll.fisher_diag_from_table_n92()
        else:
            self.ll.make_fisher_table_diag_direct_it92(s_array, d2ll_dab2)
            self.ll.fisher_diag_from_table_it92()
        #json.dump(dict(b=self.ll.table_bs, pp1=self.ll.pp1, bb=self.ll.bb),
        #          open("ll_fisher.json", "w"), indent=True)
        #a, (b,c) = ll.fisher_for_coo()
        #json.dump(([float(x) for x in a], ([int(x) for x in b], [int(x) for x in c])), open("fisher.json", "w"))
        if specs is not None:
            self.ll.spec_correction(specs)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import os
import time
import re
import gemmi
import numpy
import json
import pandas
import scipy.sparse
from dataclasses import dataclass, field
from typing import List, Dict
import omegaconf
import servalcat # for version
from servalcat.utils import logger
from servalcat import utils
from servalcat.refmac import exte
from servalcat import ext
from . import cgsolve
u_to_b = utils.model.u_to_b
b_to_u = utils.model.b_to_u
Type = ext.RefineParams.Type
#import line_profiler
#import atexit
#profile = line_profiler.LineProfiler()
#atexit.register(profile.print_stats)

"""
atom_selection:
  xyz:
    include: []
    exclude: []
    exclude_restraint: []
  adp:
    include: []
    exclude: []
    exclude_restraint: []
  occ:
    include: []
    exclude: []
    exclude_restraint: []
  dfrac:
    include: []
    exclude: []

occ_groups:
  - id: 1
    selections:
      - sel_1
      - sel_2
occ_group_constraints:
  - ids: [1, 2]
    complete: true

initialisation:
  adp:
    '*': 50
  occ: {}
  dfrac:
    '[H]': 1.0
"""

@dataclass
class SelectionConfig:
    include: List[str] = field(default_factory=list, metadata={"help": "List of gemmi Selection to include"})
    exclude: List[str] = field(default_factory=list, metadata={"help": "List of gemmi Selection to exclude"})
    exclude_restraint: List[str] = field(default_factory=list, metadata={"help": "List of gemmi Selection to exclude from restraints"})
@dataclass
class OccGroupItem:
    id: int
    selections: List[str] = field(default_factory=list)
    
@dataclass
class OccGroupConstItem:
    ids: List[int] = field(default_factory=list)
    complete: bool = True
    
@dataclass
class RefineConfig:
    atom_selection: Dict[str, SelectionConfig] = field(
        default_factory=lambda: {
            "xyz": SelectionConfig(include=["*"]),
            "adp": SelectionConfig(include=["*"]),
            "occ": SelectionConfig(),
            "dfrac": SelectionConfig(),
        },
        metadata={"help": "Configuration for atom selection during refinement"}
    )
    occ_groups: List[OccGroupItem] = field(default_factory=list)
    occ_group_constraints: List[OccGroupConstItem] = field(default_factory=list)
    occ_group_const_mu: float = 10
    occ_group_const_mu_update_factor: float = 1.1
    occ_group_const_mu_update_tol_rel: float = 0.25
    occ_group_const_mu_update_tol_abs: float = 0.01
    initialisation: Dict[str, Dict[str, float]] = field(
        default_factory=lambda: {
            "adp": {},
            "occ": {},
            "dfrac": {}
        },
        metadata={"help": ""}
    )
    write_trajectory: bool = False

def load_config(yaml_file, args, refmac_params):
    cfg = omegaconf.OmegaConf.create({"refine": RefineConfig()})
    if yaml_file:
        conf = omegaconf.OmegaConf.load(yaml_file)
        try:
            cfg = omegaconf.OmegaConf.merge(cfg, conf)
        except omegaconf.errors.ValidationError as e:
            raise SystemExit(f"Error while parsing {yaml_file}.\n{e.msg}")
            
    # load from args
    rcfg = cfg.refine
    if getattr(args, "write_trajectory", False):
        rcfg.write_trajectory = True
    if getattr(args, "fix_xyz", False):
        rcfg.atom_selection.xyz.include = []
        rcfg.atom_selection.xyz.exclude = []
    if getattr(args, "adp", None) == "fix":
        rcfg.atom_selection.adp.include = []
        rcfg.atom_selection.adp.exclude = []
    if getattr(args, "refine_all_occ", False):
        rcfg.atom_selection.occ.include = ["*"]
        rcfg.atom_selection.occ.exclude = []
    if getattr(args, "refine_dfrac", False):
        rcfg.atom_selection.dfrac.include = ["*"] # or H?
        rcfg.atom_selection.dfrac.exclude = []

    # load Refmac params (unfinished)
    if refmac_params.get("occu", {}).get("groups"):
        rgroups = refmac_params["occu"]["groups"]
        rconst = refmac_params["occu"].get("const", [])
        occ_grs = rcfg.occ_groups
        occ_cnst = rcfg.occ_group_constraints
        occ_incl = rcfg.atom_selection.occ.include
        for igr in rgroups:
            occ_grs.append(OccGroupItem(igr))
            for sel in rgroups[igr]:
                for chain in sel.get("chains", ["*"]):
                    if "resi_from" in sel and "resi_to" in sel:
                        resi = f"{sel['resi_from']}-{sel['resi_to']}"
                    else:
                        resi = sel.get("resi", "")
                    atom = sel.get("atom", "*")
                    alt = sel.get("alt", "")
                    if alt: alt = ":" + alt
                    selstr = f"//{chain}/{resi}/{atom}{alt}"
                    occ_grs[-1]["selections"].append(selstr)
                    #occ_incl.append(selstr) # to do this, GroupOccupancy needs to be removed
        for cmpl, ids in rconst:
            occ_cnst.append(OccGroupConstItem(ids, cmpl))
    
    logger.writeln("Config loaded")
    logger.writeln("--")
    logger.write(omegaconf.OmegaConf.to_yaml(cfg))
    logger.writeln("--")
    return cfg.refine
# load_config()

def RefineParams(st, refine_xyz=False, adp_mode=0, refine_occ=False,
                 refine_dfrac=False, use_q_b_mixed=True, cfg=None,
                 exclude_h_ll=True): # FIXME refine_dfrac/exclude_h_ll and cfg
    assert adp_mode in (0, 1, 2) # 0=fix, 1=iso, 2=aniso
    if refine_dfrac and not st[0].has_hydrogen():
        raise RuntimeError("Hydrogen must be present when deuterium fraction refinement is requested")
    ret = ext.RefineParams(use_aniso=(adp_mode == 2), use_q_b_mixed=use_q_b_mixed)
    ret.set_model(st[0])
    if cfg:
        # occupancy groups
        occ_groups = []
        group_ids = {}
        for occ_gr in cfg.occ_groups:
            occ_groups.append([])
            group_ids[occ_gr.id] = len(occ_groups) - 1
            for s in occ_gr.selections:
                sel = gemmi.Selection(s)
                nsel = 0
                for model in sel.models(st):
                    for chain in sel.chains(model):
                        for residue in sel.residues(chain):
                            for atom in sel.atoms(residue):
                                occ_groups[-1].append(atom)
                                nsel += 1
                if nsel == 0:
                    logger.writeln(f"Warning: no atom found for the selection {s}")
        ret.set_occ_groups(occ_groups)
        for o in cfg.occ_group_constraints:
            ret.occ_group_constraints.append((o.complete, [group_ids[x] for x in o.ids]))

        # selections
        sele = cfg.atom_selection
        ext.set_refine_flags(st[0],
                             sele.xyz.include, sele.xyz.exclude,
                             sele.adp.include, sele.adp.exclude,
                             sele.occ.include, sele.occ.exclude,
                             sele.dfrac.include, sele.dfrac.exclude)
        ret.set_params_from_flags()

        for t, p in ((Type.X, sele.xyz), (Type.B, sele.adp), (Type.Q, sele.occ)):
            for ex_sel in p.exclude_restraint:
                sel = gemmi.Selection(ex_sel)
                for model in sel.models(st):
                    for chain in sel.chains(model):
                        for residue in sel.residues(chain):
                            for atom in sel.atoms(residue):
                                ret.add_geom_exclusion(atom.serial-1, t)
    else:
        ret.set_params(refine_xyz=refine_xyz, refine_adp=adp_mode > 0,
                       refine_occ=refine_occ, refine_dfrac=refine_dfrac)
    if exclude_h_ll:
        if refine_dfrac:
            for t in (Type.X, Type.B, Type.Q):
                if ret.is_refined(t):
                    ret.exclude_h_ll(t)
        else:
            ret.exclude_h_ll()

    logger.writeln("Number of refinement parameters:")
    df = pandas.DataFrame(ret.params_summary())
    logger.writeln(df.to_string() + "\n")
    return ret

class Geom:
    def __init__(self, st, topo, monlib, refine_params, adpr_w=1, occr_w=1, shake_rms=0,
                 params=None, unrestrained=False, use_nucleus=False,
                 ncslist=None):
        self.st = st
        self.params = refine_params
        self.lookup = {x.atom: x for x in self.st[0].all()}
        try:
            self.geom = ext.Geometry(self.st, self.params, monlib.ener_lib)
        except TypeError as e:
            raise SystemExit(f"An error occurred while creating the Geometry object:\n{e}\n\n"
                             "This likely indicates an installation issue. "
                             "Please verify that you have the correct version of gemmi installed and that both gemmi and servalcat were compiled in the same environment.")
        self.specs = utils.model.find_special_positions(self.st)
        #cs_count = len(self.st.find_spacegroup().operations())
        for atom, images, matp, mata in self.specs:
            #n_sym = len([x for x in images if x < cs_count]) + 1
            n_sym = len(images) + 1
            self.geom.specials.append(ext.Geometry.Special(atom, matp, mata, n_sym))
        self.adpr_w = adpr_w
        self.occr_w = occr_w
        self.unrestrained = unrestrained
        if shake_rms > 0:
            numpy.random.seed(0)
            utils.model.shake_structure(self.st, shake_rms, copy=False)
            #utils.fileio.write_model(self.st, "shaken", pdb=True, cif=True)
        self.use_nucleus = use_nucleus
        self.calc_kwds = {"use_nucleus": self.use_nucleus}
        if params is None:
            params = {}
        for k in ("wbond", "wangle", "wtors", "wplane", "wchir", "wvdw", "wncs"):
            if k in params:
                self.calc_kwds[k] = params[k]
                logger.writeln("setting geometry weight {}= {}".format(k, params[k]))
        inc_tors, exc_tors = utils.restraints.make_torsion_rules(params.get("restr", {}))
        rtors = utils.restraints.select_restrained_torsions(monlib, inc_tors, exc_tors)
        self.geom.mon_tors_names = rtors["monomer"]
        self.geom.link_tors_names = rtors["link"]
        self.group_occ = GroupOccupancy(self.st, params.get("occu"))
        if not self.unrestrained:
            self.geom.load_topo(topo)
        exte.read_external_restraints(params.get("exte", []), self.st, self.geom)
        self.geom.finalize_restraints()
        self.outlier_sigmas = dict(bond=5, angle=5, torsion=5, vdw=5, ncs=5, chir=5, plane=5, staca=5, stacd=5, per_atom=5, interval=5)
        self.parents = {}
        self.ncslist = ncslist
        self.const_ls, self.const_u = [], []
    # __init__()

    def set_h_parents(self):
        self.parents = {}
        for bond in self.geom.bonds:
            if bond.atoms[0].is_hydrogen():
                self.parents[bond.atoms[0]] = bond.atoms[1]
            elif bond.atoms[1].is_hydrogen():
                self.parents[bond.atoms[1]] = bond.atoms[0]
    # set_h_parents()

    def setup_target(self):
        self.geom.setup_target(self.params.is_refined(Type.Q))
    def setup_nonbonded(self):
        skip_critical_dist = not self.params.is_refined(Type.X) or self.unrestrained
        self.geom.setup_nonbonded(skip_critical_dist=skip_critical_dist)
        if self.ncslist:
            self.geom.setup_ncsr(self.ncslist)
    def setup_occ_constraint(self, lambda_ini=0., u_ini=100.):
        self.const_ls = [lambda_ini for _ in self.params.occ_group_constraints]
        self.const_u = [u_ini for _ in self.params.occ_group_constraints]
    def calc(self, target_only):
        if self.params.is_refined(Type.X) and not self.unrestrained:
            return self.geom.calc(check_only=target_only, **self.calc_kwds)
        return 0
    def calc_adp_restraint(self, target_only):
        if self.params.is_refined(Type.B):
            return self.geom.calc_adp_restraint(target_only, self.adpr_w)
        return 0
    def calc_occ_restraint(self, target_only):
        if self.params.is_refined(Type.Q):
            return self.geom.calc_occ_restraint(target_only, self.occr_w)
        return 0
    def update_occ_consts(self, consts_prev, alpha=1.1, eta=0.25, tol=0.01):
        consts = self.params.occ_constraints()
        self.const_ls = [self.const_ls[i] - self.const_u[i] * consts[i]
                         for i in range(len(consts))]
        self.const_u = [u * (1 if abs(c) < max(tol, eta * abs(c_prev)) else alpha)
                        for u, c, c_prev in zip(self.const_u, consts, consts_prev)]
        return consts
    def calc_target(self, target_only):
        self.geom.clear_target()
        geom_x = self.calc(target_only) 
        geom_a = self.calc_adp_restraint(target_only)
        geom_q = self.calc_occ_restraint(target_only)
        geom_c = self.geom.calc_occ_constraint(target_only, self.const_ls, self.const_u)
        logger.writeln(" geom_x = {}".format(geom_x))
        logger.writeln(" geom_a = {}".format(geom_a))
        logger.writeln(" geom_q = {}".format(geom_q))
        logger.writeln(" geom_c = {}".format(geom_c))
        geom = geom_x + geom_a + geom_q + geom_c
        if not target_only:
            self.geom.spec_correction()
        return geom
        
    def show_model_stats(self, show_outliers=True):
        self.calc(True)
        self.calc_adp_restraint(True)
        self.calc_occ_restraint(True)
        ret = {"outliers": {}}
        if show_outliers:
            get_table = dict(bond=self.geom.reporting.get_bond_outliers,
                             angle=self.geom.reporting.get_angle_outliers,
                             torsion=self.geom.reporting.get_torsion_outliers,
                             chir=self.geom.reporting.get_chiral_outliers,
                             plane=self.geom.reporting.get_plane_outliers,
                             staca=self.geom.reporting.get_stacking_angle_outliers,
                             stacd=self.geom.reporting.get_stacking_dist_outliers,
                             vdw=self.geom.reporting.get_vdw_outliers,
                             interval=self.geom.reporting.get_interval_outliers,
                             #ncs=self.geom.reporting.get_ncsr_outliers, # not useful?
                             )
            labs = dict(bond="Bond distances",
                        angle="Bond angles",
                        torsion="Torsion angles",
                        chir="Chiral centres",
                        plane="Planar groups",
                        staca="Stacking plane angles",
                        stacd="Stacking plane distances",
                        vdw="VDW repulsions",
                        interval="Interval",
                        ncs="Local NCS restraints")

            def atomlabel(r, i):
                symstr = lambda idx, s: f" ({idx+1};{s[0]},{s[1]},{s[2]})"
                ret = str(self.lookup[r.atoms[i]])
                if type(r) in (ext.Geometry.Bond, ext.Geometry.Interval, ext.Geometry.Vdw) and i > 0 and not r.same_asu():
                    return ret + symstr(r.sym_idx, r.pbc_shift)
                if type(r) is ext.Geometry.Angle and not r.same_asu(i): # always true if i==1
                    return ret + symstr(r.sym_idx_1 if i == 0 else r.sym_idx_2,
                                        r.pbc_shift_1 if i == 0 else r.pbc_shift_2)
                return ret
            
            for k in get_table:
                kwgs = {"min_z": self.outlier_sigmas[k]}
                if k == "bond": kwgs["use_nucleus"] = self.use_nucleus
                table = get_table[k](**kwgs)
                if table["z"]:
                    if "restr" in table:
                        tmp = {}
                        for i in range(3 if k == "angle" else 2): # only bond/angle/interval/vdw return restr
                            tmp[f"atom{i+1}"] = [atomlabel(r, i) for r in table["restr"]]
                        del table["restr"]
                        table = {**tmp, **table}
                    else:
                        for kk in table:
                            if kk.startswith(("atom", "plane", "1_atom", "2_atom")):
                                table[kk] = [str(self.lookup[x]) for x in table[kk]]
                    df = pandas.DataFrame(table)
                    df = df.reindex(df.z.abs().sort_values(ascending=False).index)
                    ret["outliers"][k] = df
                    if k == "bond":
                        df0 = df[df.type < 2].drop(columns=["type", "alpha"])
                        if len(df0.index) > 0:
                            logger.writeln(" *** {} outliers (Z >= {}) ***\n".format(labs[k], self.outlier_sigmas[k]))
                            logger.writeln(df0.to_string(float_format="{:.3f}".format, index=False) + "\n")
                        df0 = df[df.type == 2].drop(columns=["type"])
                        if len(df0.index) > 0:
                            logger.writeln(" *** External bond outliers (Z >= {}) ***\n".format(self.outlier_sigmas[k]))
                            logger.writeln(df0.to_string(float_format="{:.3f}".format, index=False) + "\n")
                    else:
                        logger.writeln(" *** {} outliers (Z >= {}) ***\n".format(labs[k], self.outlier_sigmas[k]))
                        logger.writeln(df.to_string(float_format="{:.3f}".format, index=False) + "\n")

        # Per-atom score
        if 0:
            peratom = self.geom.reporting.per_atom_score(len(self.atoms), self.use_nucleus, "mean")
            df = pandas.DataFrame(peratom)
            df.insert(0, "atom", [str(self.lookup[x]) for x in self.atoms])
            df = df[df["total"] >= self.outlier_sigmas["per_atom"]]
            if show_outliers and len(df.index) > 0:
                df.sort_values("total", ascending=False, inplace=True)
                ret["outliers"]["per_atom"] = df
                logger.writeln(" *** Per-atom violations (Z >= {}) ***\n".format(self.outlier_sigmas["per_atom"]))
                logger.writeln(df.to_string(float_format="{:.2f}".format, index=False) + "\n")

        df = pandas.DataFrame(self.geom.reporting.get_summary_table(self.use_nucleus))
        df = df.set_index("Restraint type").rename_axis(index=None)
        ret["summary"] = df
        logger.writeln(df.to_string(float_format="{:.3f}".format) + "\n")
        return ret

def show_binstats(df, cycle_number):
    forplot = []
    datalabs = [x for x in ("Mn(Io)", "Mn(Ic)", "Mn(Fo)", "Mn(Fc)") if x in df]
    rlabs = [x for x in df if x.startswith("R")]
    fsclabs = [x for x in df if x.startswith("fsc")]
    cclabs = [x for x in df if x.startswith("CC")]
    dlabs = [x for x in df if re.search("^D[0-9]*", x)]
    if datalabs: forplot.append(["Mean I/F vs. Resolution", datalabs])
    if "fsc_model" in df: forplot.append(["FSC", ["fsc_model"]])
    if rlabs: forplot.append(["R", rlabs])
    if fsclabs: forplot.append(["FSC", fsclabs])
    if cclabs: forplot.append(["CC", cclabs])
    if dlabs: forplot.append(["ML parameters - D", dlabs])
    if "S" in df: forplot.append(["ML parameters - Sigma", ["S"]])
    if "Cmpl" in df: forplot.append(["Data completeness", ["Cmpl"]])
    lstr = utils.make_loggraph_str(df, "Data stats in cycle {}".format(cycle_number), forplot,
                                   s2=1/df["d_min"]**2,
                                   float_format="{:.4f}".format)
    logger.writeln(lstr)
# show_binstats()

def convert_stats_to_dicts(stats):
    tmp = []
    for s in stats: # stats must be a list of dict
        tmp.append({})
        for k in s:
            if k == "geom":
                tmp[-1]["geom"] = {"summary": s["geom"]["summary"].to_dict()}
                for kk in s["geom"]["outliers"]:
                    tmp[-1]["geom"].setdefault("outliers", {})[kk] = s["geom"]["outliers"][kk].to_dict(orient="records")
            else:
                tmp[-1][k] = s[k]
    return tmp
# convert_stats_to_dicts()

def write_stats_json_safe(stats, json_out):
    tmp = convert_stats_to_dicts(stats)
    out_tmp = json_out + ".part"
    with open(out_tmp, "w") as ofs:
        json.dump(tmp, ofs, indent=2)
    for i in range(10):
        try:
            # On Windows, this fails when another process open the file
            os.replace(out_tmp, json_out)
            break
        except PermissionError:
            logger.writeln(f"{json_out} locked. retrying..")
            time.sleep(0.5)
    else:
        raise RuntimeError(f"Cannot write {json_out}")
    logger.writeln(f"Refinement statistics saved: {json_out}")
# write_stats_json_safe()

def print_h_options(h_change, h_present, refine_h, hout, geom_only):
    if not h_present:
        h_change = gemmi.HydrogenChange.Remove
    logger.writeln("Hydrogen related options")
    logger.write(" use in refinement{}: hydrogen atoms ".format("" if geom_only else "/map calculation"))
    logger.writeln({gemmi.HydrogenChange.ReAddButWater: "have been (re)generated",
                    gemmi.HydrogenChange.ReAdd:         "(including water) have been (re)generated",
                    gemmi.HydrogenChange.ReAddKnown:    "(except for rotatable) have been (re) generated",
                    gemmi.HydrogenChange.NoChange:      "from the input model have been retained",
                    gemmi.HydrogenChange.Remove:        "have either been removed or were not present"}[h_change])
    if h_present:
        logger.write(" target: hydrogen atoms will be ")
        if geom_only or not refine_h:
            logger.writeln("just optimized according to geometric restraints")
        else:
            logger.writeln("refined against experimental data")
    logger.writeln(" in output model: " + ("written" if hout and h_present else "not written"))
    logger.writeln("")
# print_hydrogen_options()

class GroupOccupancy:
    # TODO max may not be one. should check multiplicity
    def __init__(self, st, params):
        self.groups = []
        self.consts = []
        self.group_idxes = [0 for _ in range(st[0].count_atom_sites())]
        self.ncycle = 0
        if not params or not params.get("groups"):
            return
        logger.writeln("Occupancy groups:")
        atom_idxes = []
        for igr in params["groups"]:
            self.groups.append([]) # list of atoms
            n_curr = len(atom_idxes)
            for sel in params["groups"][igr]:
                sel_chains = sel.get("chains")
                sel_from = sel.get("resi_from")
                sel_to = sel.get("resi_to")
                sel_seq = sel.get("resi")
                sel_atom = sel.get("atom")
                sel_alt = sel.get("alt")
                for chain in st[0]:
                    if sel_chains and chain.name not in sel_chains:
                        continue
                    flag = False
                    for res in chain:
                        if sel_seq and res.seqid != sel_seq:
                            continue
                        if sel_from and res.seqid == sel_from:
                            flag = True
                        if sel_from and not flag:
                            continue
                        for atom in res:
                            if sel_atom and atom.name != sel_atom:
                                continue
                            if sel_alt and atom.altloc != sel_alt:
                                continue
                            atom_idxes.append(atom.serial-1)
                            self.groups[-1].append(atom)
                            self.group_idxes[atom.serial-1] = len(self.groups)
                        if sel_to and res.seqid == sel_to:
                            flag = False
            logger.writeln(" id= {} atoms= {}".format(igr, len(atom_idxes) - n_curr))

        igr_idxes = {igr:i for i, igr in enumerate(params["groups"])}
        self.consts = [(is_comp, [igr_idxes[g] for g in gids])
                       for is_comp, gids in params["const"]]
        self.ncycle = params.get("ncycle", 5)
        self.params = ext.RefineParams()
        self.params.set_model(st[0])
        self.params.set_params_selected(atom_idxes, refine_occ=True)
        self.params.exclude_h_ll() # should be reasonable
    # __init__()
    
    def constraint(self, x):
        # x: occupancy parameters
        ret = []
        for is_comp, ids in self.consts:
            x_sum = numpy.sum(x[ids])
            if is_comp or x_sum > 1:
                ret.append(x_sum - 1)
            else:
                ret.append(0.)
        return numpy.array(ret)
        
    def ensure_constraints(self):
        vals = []
        for atoms in self.groups:
            occ = numpy.mean([a.occ for a in atoms])
            occ = min(1, max(1e-3, occ))
            vals.append(occ)
        for is_comp, idxes in self.consts:
            sum_occ = sum(vals[i] for i in idxes)
            if not is_comp and sum_occ < 1:
                sum_occ = 1. # do nothing
            for i in idxes:
                logger.writeln("Imposing constraints: {} {}".format(vals[i], vals[i]/sum_occ))
                vals[i] /= sum_occ
        for occ, atoms in zip(vals, self.groups):
            for a in atoms: a.occ = occ
        
    def get_x(self):
        return numpy.array([atoms[0].occ for atoms in self.groups])

    def set_x(self, x):
        for p, atoms in zip(x, self.groups):
            for a in atoms:
                a.occ = p
                #a.occ = max(1, min(1e-3, p))

    def target(self, x, ll, ls, u):
        self.set_x(x)
        ll.update_fc()
        c = self.constraint(x)
        f = ll.calc_target() - numpy.dot(ls, c) + 0.5 * u * numpy.sum(c**2)
        return f
    
    def grad(self, x, ll, ls, u):
        c = self.constraint(x)
        ll.calc_grad(self.params, specs=None)
        #print("grad=", ll.ll.vn)
        #print("diag=", ll.ll.am)
        assert len(ll.ll.vn) == len(ll.ll.am)
        vn = []
        diag = []
        atom_to_param = self.params.atom_to_param(Type.Q)
        for atoms in self.groups: # idxes
            idxes = [atom_to_param[a.serial-1] for a in atoms if not self.params.is_excluded_ll(a.serial-1, Type.Q)]
            vn.append(numpy.sum(numpy.array(ll.ll.vn)[idxes]))
            diag.append(numpy.sum(numpy.array(ll.ll.am)[idxes]))
        vn, diag = numpy.array(vn), numpy.array(diag)
        for i, (is_comp, idxes) in enumerate(self.consts):
            dcdx = numpy.zeros(len(self.groups))
            dcdx[idxes] = 1.
            if is_comp or c[i] != 0:
                vn -= (ls[i] - u * c[i]) * dcdx
            diag += u * dcdx**2

        return vn, diag
        
    def refine(self, ll, alpha=1.1):
        # Refinement of grouped occupancies using augmented Lagrangian
        # f(x) = LL(x) - sum_j (lambda_j c_j(x)) + u/2 sum_j (c_j(x))^2
        # with c_j(x) = 0 constraints
        if not self.groups:
            return
        logger.writeln("\n== Group occupancy refinement ==")
        self.ensure_constraints() # make sure constrained groups have the same occupancies.
        ls = 0 * numpy.ones(len(self.consts)) # Lagrange multiplier
        u = 10000. # penalty parameter. in Refmac 1/0.01**2
        x0 = self.get_x()
        #logger.writeln("  parameters: {}".format(len(x0)))
        f0 = self.target(x0, ll, ls, u)
        ret = []
        for cyc in range(self.ncycle):
            ret.append({"Ncyc": cyc+1, "f0": f0})
            logger.writeln("occ_{}_f0= {:.4e}".format(cyc, f0))
            vn, diag = self.grad(x0, ll, ls, u)
            #diag0 = diag.copy()
            diag[diag < 1e-6] = 1.
            dx = -vn / diag
            #logger.writeln(f"debug {cyc=} {dx=} {vn=} {diag=} {diag0=}")
            if 0:
                ofs = open("debug.dat", "w")
                for scale in (-1, -0.5, 0, 0.1, 0.2, 0.3, 0.4, 0.5, 1, 2):
                    self.set_x(x0 + scale * dx)
                    ll.update_fc()
                    c = self.constraint(x0 + dx)
                    f = ll.calc_target() + numpy.dot(ls, c) + 0.5 * u * numpy.sum(c**2)
                    ofs.write("{} {}\n".format(scale, f))
                ofs.close()
                import scipy.optimize
                print(scipy.optimize.line_search(f=lambda x: self.target(x, ll, ls, u),
                                                 myfprime= lambda x: self.grad(ll, ls, u)[0],
                                                 xk= x0,
                                                 pk= dx))
                quit()

            scale = 1
            for i in range(3):
                scale = 1/2**i
                f1 = self.target(x0 + dx * scale, ll, ls, u)
                logger.writeln("occ_{}_f1, {}= {:.4e}".format(cyc, i, f1))
                if f1 < f0: break
            else:
                logger.writeln("WARNING: function not minimised")
                #self.set_x(x0) # Refmac accepts it even when function increases
            c = self.constraint(x0 + dx * scale)
            ret[-1]["f1"] = f1
            ret[-1]["shift_scale"] = scale
            f0 = f1
            x0 = x0 + dx * scale
            ls -= u * c
            u = alpha * u
            ret[-1]["const_viol"] = list(c)
            ret[-1]["lambda_new"] = list(ls)
        self.ensure_constraints()
        ll.update_fc()
        f = ll.calc_target()
        logger.writeln("final -LL= {}".format(f))
        return ret


class Refine:
    def __init__(self, st, geom, cfg, refine_params, ll=None, unrestrained=False):
        assert geom is not None
        self.st = st # clone()?
        self.st_traj = None
        self.params = refine_params
        self.geom = geom
        self.ll = ll
        self.gamma = 0
        self.unrestrained = unrestrained
        #self.h_inherit_parent_adp = self.params.is_refined(Type.B) and not self.refine_h and self.st[0].has_hydrogen()
        #if self.h_inherit_parent_adp:
        #    self.geom.set_h_parents()
        self.cfg = cfg
        if self.cfg.write_trajectory:
            self.st_traj = self.st.clone()
            self.st_traj[-1].num = 0
        assert self.geom.group_occ.groups or self.params.n_params() > 0
    # __init__()
    
    def print_weights(self): # TODO unfinished
        logger.writeln("Geometry weights")
        g = self.geom.geom
        if self.params.is_refined(Type.B):
            logger.writeln(" ADP restraints")
            logger.writeln("  weight: {}".format(self.geom.adpr_w))
            logger.writeln("  mode: {}".format(g.adpr_mode))
            if g.adpr_mode == "diff":
                logger.writeln("  sigmas: {}".format(" ".join("{:.2f}".format(x) for x in g.adpr_diff_sigs)))
            elif g.adpr_mode == "kldiv":
                logger.writeln("  sigmas: {}".format(" ".join("{:.2f}".format(x) for x in g.adpr_kl_sigs)))
            else:
                raise LookupError("unknown adpr_mode")
        if self.params.is_refined(Type.Q):
            logger.writeln(" Occupancy restraints")
            logger.writeln("  weight: {}".format(self.geom.occr_w))

    def scale_shifts(self, dx, scale):
        shift_allow_high =  1.0
        shift_allow_low  = -1.0
        shift_max_allow_B = 30.0
        shift_min_allow_B = -30.0
        shift_max_allow_q = 0.5
        shift_min_allow_q = -0.5
        shift_max_allow_d = 0.5
        shift_min_allow_d = -0.5
        dx = scale * dx
        dxx = dx[self.params.vec_selection(Type.X)]
        dxb = dx[self.params.vec_selection(Type.B)]
        dxq = dx[self.params.vec_selection(Type.Q)]
        dxd = dx[self.params.vec_selection(Type.D)]
        if len(dxx) > 0:
            logger.writeln("min(dx) = {}".format(numpy.min(dxx)))
            logger.writeln("max(dx) = {}".format(numpy.max(dxx)))
            logger.writeln("mean(dx)= {}".format(numpy.mean(dxx)))
            dxx[dxx > shift_allow_high] = shift_allow_high
            dxx[dxx < shift_allow_low] = shift_allow_low
        if len(dxb) > 0:
            # TODO this is misleading in anisotropic case
            logger.writeln("min(dB) = {}".format(numpy.min(dxb)))
            logger.writeln("max(dB) = {}".format(numpy.max(dxb)))
            logger.writeln("mean(dB)= {}".format(numpy.mean(dxb)))
            # FIXME we should'nt apply eigen decomp to dxb
            if self.params.aniso:
                for i in range(len(dxb)//6):
                    j = i * 6
                    a = numpy.array([[dxb[j],   dxb[j+3], dxb[j+4]],
                                     [dxb[j+3], dxb[j+1], dxb[j+5]],
                                     [dxb[j+4], dxb[j+5], dxb[j+2]]])
                    v, Q = numpy.linalg.eigh(a)
                    v[v > shift_max_allow_B] = shift_max_allow_B
                    v[v < shift_min_allow_B] = shift_min_allow_B
                    a = Q.dot(numpy.diag(v)).dot(Q.T)
                    dxb[j:j+6] = a[0,0], a[1,1], a[2,2], a[0,1], a[0,2], a[1,2]
            else:
                dxb[dxb > shift_max_allow_B] = shift_max_allow_B
                dxb[dxb < shift_min_allow_B] = shift_min_allow_B
        if len(dxq) > 0:
            logger.writeln("min(dq) = {}".format(numpy.min(dxq)))
            logger.writeln("max(dq) = {}".format(numpy.max(dxq)))
            logger.writeln("mean(dq)= {}".format(numpy.mean(dxq)))
            dxq[dxq > shift_max_allow_q] = shift_max_allow_q
            dxq[dxq < shift_min_allow_q] = shift_min_allow_q
        if len(dxd) > 0:
            logger.writeln("min(dd) = {}".format(numpy.min(dxd)))
            logger.writeln("max(dd) = {}".format(numpy.max(dxd)))
            logger.writeln("mean(dd)= {}".format(numpy.mean(dxd)))
            dxd[dxd > shift_max_allow_d] = shift_max_allow_d
            dxd[dxd < shift_min_allow_d] = shift_min_allow_d

        return dx

    def set_x(self, x):
        self.params.set_x(x, min_b=0.5)
        max_occ = {}
        if self.params.is_refined(Type.Q) and self.geom.specs:
            max_occ = {atom: 1./(len(images)+1) for atom, images, _, _ in self.geom.specs}
        for a in self.params.atoms:
            a.occ = min(max_occ.get(a, 1), max(0, a.occ))
        # Copy B of hydrogen from parent
        #if self.h_inherit_parent_adp:
        #    for h in self.geom.parents:
        #        p = self.geom.parents[h]
        #        h.b_iso = p.b_iso
        #        h.aniso = p.aniso

        if self.ll is not None:
            self.ll.update_fc()
        
        self.geom.setup_nonbonded() # if refine_xyz=False, no need to do it every time
        self.geom.setup_target()
        logger.writeln("vdws = {}".format(len(self.geom.geom.vdws)))
        logger.writeln(f"atoms = {len(self.params.atoms)}")
        logger.writeln(f"pairs = {self.geom.geom.target.n_pairs()}")

    def get_x(self):
        return numpy.array(self.params.get_x())
    
    #@profile
    def calc_target(self, w=1, target_only=False):
        geom = self.geom.calc_target(target_only)
        if self.ll is not None:
            ll = self.ll.calc_target()
            logger.writeln(" ll= {}".format(ll))
            if not target_only:
                self.ll.calc_grad(self.params, self.geom.geom.specials)
        else:
            ll = 0

        f =  w * ll + geom
        return f

    #@profile
    def run_cycle(self, weight=1):
        if 0: # test of grad
            self.ll.update_fc()
            x0 = self.get_x()
            f0,ader,_ = self.calc_target(weight)
            i = 1
            for e in 1e-1,1e-2,1e-3, 1e-4, 1e-5:
                x1 = numpy.copy(x0)
                x1[i] += e
                self.set_x(x1)
                self.ll.update_fc()
                f1,_,_ = self.calc_target(weight, target_only=True)
                nder = (f1 - f0) / e
                print("e=", e)
                print("NUM DER=", nder)
                print("ANA DER=", ader[i])
                print("ratio=", nder/ader[i])
            quit()

        f0 = self.calc_target(weight)
        x0 = self.get_x()
        logger.writeln("f0= {:.4e}".format(f0))
        if 0:
            logger.writeln(f"geom_vec=\n{self.geom.geom.target.vn}")
            logger.writeln(f"geom_mat=\n{self.geom.geom.target.am_spmat}")
            logger.writeln(f"  ll_vec=\n{self.ll.ll.vn}")
            logger.writeln(f"  ll_mat=\n{self.ll.ll.fisher_spmat}")
        if 1:
            use_ic = False # incomplete cholesky. problematic at least in geometry optimisation case
            logger.writeln("using cgsolve in c++, ic={}".format(use_ic))
            cgsolver = ext.CgSolve(self.geom.geom.target, None if self.ll is None else self.ll.ll)
            if use_ic:
                cgsolver.gamma = 0
                cgsolver.max_gamma_cyc = 1
            else:
                cgsolver.gamma = self.gamma
            dx = cgsolver.solve(weight, logger, use_ic)
            self.gamma = cgsolver.gamma
        else:
            logger.writeln("using cgsolve in py")
            am = self.geom.geom.target.am_spmat
            vn = numpy.array(self.geom.geom.target.vn)
            if self.ll is not None:
                am += self.ll.ll.fisher_spmat * weight
                vn += numpy.array(self.ll.ll.vn) * weight
            diag = am.diagonal()
            diag[diag<=0] = 1.
            diag = numpy.sqrt(diag)
            rdiag = 1./diag # sk
            M = scipy.sparse.diags(rdiag)
            dx, self.gamma = cgsolve.cgsolve_rm(A=am, v=vn, M=M, gamma=self.gamma)

        if 0: # to check hessian scale
            with open("minimise_line.dat", "w") as ofs:
                ofs.write("s f\n")
                for s in numpy.arange(-2, 2, 0.1):
                    dx2 = self.scale_shifts(dx, s)
                    self.set_x(x0 + dx2)
                    fval = self.calc_target(weight, target_only=True)
                    ofs.write("{} {}\n".format(s, fval))
            quit()

        ret = True # success
        shift_scale = 1
        for i in range(3):
            shift_scale = 1/2**i
            dx2 = self.scale_shifts(dx, shift_scale)
            self.set_x(x0 - dx2)
            f1 = self.calc_target(weight, target_only=True)
            logger.writeln("f1, {}= {:.4e}".format(i, f1))
            if f1 < f0: break
        else:
            ret = False
            logger.writeln("WARNING: function not minimised")
            #self.set_x(x0) # Refmac accepts it even when function increases

        return ret, shift_scale, f1

    def run_cycles(self, ncycles, weight=1, weight_adjust=False, debug=False,
                   weight_adjust_bond_rmsz_range=(0.5, 1.), stats_json_out=None):
        self.print_weights()
        stats = [{"Ncyc": 0}]
        self.params.ensure_occ_constraints()
        self.geom.setup_nonbonded()
        self.geom.setup_target()
        self.geom.setup_occ_constraint(u_ini=self.cfg.occ_group_const_mu)
        logger.writeln("vdws = {}".format(len(self.geom.geom.vdws)))
        logger.writeln(f"atoms = {len(self.params.atoms)}")
        logger.writeln(f"pairs = {self.geom.geom.target.n_pairs()}")
        stats[-1]["geom"] = self.geom.show_model_stats(show_outliers=True)
        if self.params.occ_group_constraints:
            stats[-1]["occ_const"] = {"lambda": self.geom.const_ls,
                                      "mu": self.geom.const_u,
                                      "violation": self.params.occ_constraints(),
                                      "occ": self.params.constrained_occ_values()
                                      }
        if self.ll is not None:
            self.ll.update_fc()
            self.ll.overall_scale()
            self.ll.update_ml_params()
            self.ll.prepare_target()
            llstats = self.ll.calc_stats(bin_stats=True)
            stats[-1]["data"] = {"summary": llstats["summary"],
                                 "binned": llstats["bin_stats"].to_dict(orient="records"),
                                 "ml": llstats["ml"].to_dict(orient="records")}
            if "twin_alpha" in llstats:
                stats[-1]["twin_alpha"] = llstats["twin_alpha"]
            show_binstats(llstats["bin_stats"], 0)
        if self.params.is_refined(Type.B):
            utils.model.adp_analysis(self.st)
        if stats_json_out:
            write_stats_json_safe(stats, stats_json_out)
        occ_refine_flag = self.ll is not None and self.geom.group_occ.groups and self.geom.group_occ.ncycle > 0

        for i in range(ncycles):
            logger.writeln("\n====== CYCLE {:2d} ======\n".format(i+1))
            logger.writeln(f" weight = {weight:.4e}")
            if self.params.is_refined_any():
                is_ok, shift_scale, fval = self.run_cycle(weight=weight)
                stats.append({"Ncyc": len(stats), "shift_scale": shift_scale, "fval": fval, "fval_decreased": is_ok,
                              "weight": weight})
            elif occ_refine_flag:
                stats.append({"Ncyc": len(stats)})
            if occ_refine_flag:
                stats[-1]["occ_refine"] = self.geom.group_occ.refine(self.ll)
            if debug: utils.fileio.write_model(self.st, "refined_{:02d}".format(i+1), pdb=True)#, cif=True)
            stats[-1]["geom"] = self.geom.show_model_stats(show_outliers=(i==ncycles-1))
            if self.params.occ_group_constraints:
                viols = self.geom.update_occ_consts(consts_prev=stats[-2]["occ_const"]["violation"],
                                                    alpha=self.cfg.occ_group_const_mu_update_factor,
                                                    eta=self.cfg.occ_group_const_mu_update_tol_rel,
                                                    tol=self.cfg.occ_group_const_mu_update_tol_abs)
                stats[-1]["occ_const"] = {"lambda": self.geom.const_ls,
                                          "mu": self.geom.const_u,
                                          "violation": viols,
                                          "occ": self.params.constrained_occ_values()
                                          }
            # TODO add stats[-1]["occ_constraints"] and hide stdout
            if self.ll is not None:
                if i == ncycles - 1: # last cycle
                    self.params.ensure_occ_constraints()
                self.ll.overall_scale()
                f0 = self.ll.calc_target()
                self.ll.update_ml_params()
                self.ll.prepare_target()
                llstats = self.ll.calc_stats(bin_stats=True)#(i==ncycles-1))
                if llstats["summary"]["-LL"] > f0:
                    logger.writeln("WARNING: -LL has increased after ML parameter optimization:"
                                   "{} to {}".format(f0, llstats["summary"]["-LL"]))
                stats[-1]["data"] = {"summary": llstats["summary"],
                                     "binned": llstats["bin_stats"].to_dict(orient="records"),
                                     "ml": llstats["ml"].to_dict(orient="records")}
                if "twin_alpha" in llstats:
                    stats[-1]["twin_alpha"] = llstats["twin_alpha"]
                show_binstats(llstats["bin_stats"], i+1)
            if self.params.is_refined(Type.B):
                utils.model.adp_analysis(self.st)
            if (weight_adjust and self.params.is_refined(Type.X) and not self.unrestrained and self.ll is not None and
                len(stats) > 2 and "Bond distances, non H" in stats[-1]["geom"]["summary"].index):
                rmsz = stats[-1]["geom"]["summary"]["r.m.s.Z"]["Bond distances, non H"]
                rmsz0 = stats[-2]["geom"]["summary"]["r.m.s.Z"]["Bond distances, non H"]
                if rmsz > weight_adjust_bond_rmsz_range[1] and rmsz > rmsz0:
                    weight /= 1.1
                elif rmsz < weight_adjust_bond_rmsz_range[0] and rmsz0 < weight_adjust_bond_rmsz_range[0] and rmsz < rmsz0:
                    weight *= 1.3
                elif rmsz > 1.5 * rmsz0:
                    weight /= 1.1
            if self.st_traj is not None:
                self.st_traj.add_model(self.st[0])
                self.st_traj[-1].num = len(self.st_traj) - 1
            if stats_json_out:
                write_stats_json_safe(stats, stats_json_out)

            logger.writeln("")

        # Make tables
        if self.params.occ_group_constraints:
            tmp = []
            for icyc, s in enumerate(stats):
                con = s["occ_const"]
                d = {"Ncyc": icyc}
                d.update({f"lambda_{i+1}":l for i,l in enumerate(con["lambda"])})
                d.update({f"mu_{i+1}":l for i,l in enumerate(con["mu"])})
                d.update({f"violation_{i+1}":l for i,l in enumerate(con["violation"])})
                d.update({f"occ_{i+1}_{j+1}":q for i, l in enumerate(con["occ"])
                          for j, q in enumerate(l)})
                tmp.append(d)
            df = pandas.DataFrame(tmp)
            forplot = [
                ["Lagrange multiplier", 
                 ["Ncyc"] + [x for x in df if x.startswith("lambda_")]],
                ["Penalty parameter mu",
                 ["Ncyc"] + [x for x in df if x.startswith("mu")]],
                ["Group constrained occupancies",
                 ["Ncyc"] + [x for x in df if x.startswith("occ_")]],
                ["Constraint violations",
                 ["Ncyc"] + [x for x in df if x.startswith("violation_")]],
            ]
            lstr = utils.make_loggraph_str(df, "group occupancies vs cycle", forplot,
                                           float_format="{:.4f}".format)
            logger.writeln(lstr)
            
        data_keys, geom_keys = set(), set()
        tmp = []
        for d in stats:
            x = {"Ncyc": d["Ncyc"]}
            if "data" in d and "summary" in d["data"]:
                x.update(d["data"]["summary"])
                data_keys.update(d["data"]["summary"])
            if "geom" in d:
                for k, n, l in (("r.m.s.d.", "Bond distances, non H", "rmsBOND"),
                                ("r.m.s.Z", "Bond distances, non H", "zBOND"),
                                ("r.m.s.d.", "Bond angles, non H", "rmsANGL"),
                                ("r.m.s.Z", "Bond angles, non H", "zANGL")):
                    if k in d["geom"]["summary"] and n in d["geom"]["summary"][k]:
                        x[l] = d["geom"]["summary"][k].get(n)
                        geom_keys.add(l)
            tmp.append(x)
        df = pandas.DataFrame(tmp)
        forplot = []
        if "FSCaverage" in data_keys:
            forplot.append(["FSC", ["Ncyc", "FSCaverage"]])
        r_keys = [x for x in data_keys if x.startswith("R")]
        if r_keys:
            forplot.append(["R", ["Ncyc"] + r_keys])
        cc_keys = [x for x in data_keys if x.startswith("CC")]
        if cc_keys:
            forplot.append(["CC", ["Ncyc"] + cc_keys])
        if "-LL" in data_keys:
            forplot.append(["-LL", ["Ncyc", "-LL"]])
        rms_keys = [x for x in geom_keys if x.startswith("rms")]
        if rms_keys:
            forplot.append(["Geometry", ["Ncyc"] + rms_keys])
        z_keys = [x for x in geom_keys if x.startswith("z")]
        if z_keys:
            forplot.append(["Geometry Z", ["Ncyc"] + z_keys])

        lstr = utils.make_loggraph_str(df, "stats vs cycle", forplot,
                                       float_format="{:.4f}".format)
        logger.writeln(lstr)
        return stats

# class Refine

def update_meta(st, stats, ll=None):
    # TODO write stats. probably geom.reporting.get_summary_table should return with _refine_ls_restr.type names
    # should remove st.mod_residues?
    st.helices.clear()
    st.sheets.clear()
    raw_remarks = [f'REMARK   3',
                   f'REMARK   3 REFINEMENT.',
                   f'REMARK   3   PROGRAM     : SERVALCAT {servalcat.__version__}',
                   f'REMARK   3   AUTHORS     : YAMASHITA,MURSHUDOV',
                   f'REMARK   3',
                   ]
    si = gemmi.SoftwareItem()
    si.classification = gemmi.SoftwareItem.Classification.Refinement
    si.name = "Servalcat"
    si.version = servalcat.__version__
    si.date = servalcat.__date__
    st.meta.software = [si]

    ri = gemmi.RefinementInfo()
    if "geom" in stats:
        restr_stats = []
        raw_remarks.append("REMARK   3  RMS DEVIATIONS FROM IDEAL VALUES        COUNT    RMS    WEIGHT")
        for k, n, l, pl in (("r.m.s.d.", "Bond distances, non H", "s_bond_nonh_d",             "BOND LENGTHS REFINED ATOMS        (A)"),
                            ("r.m.s.d.", "Bond angles, non H", "s_angle_nonh_deg",             "BOND ANGLES REFINED ATOMS   (DEGREES)"),
                            ("r.m.s.d.", "Torsion angles, period 1", "s_dihedral_angle_1_deg", "TORSION ANGLES, PERIOD 1    (DEGREES)"),
                            ("r.m.s.d.", "Torsion angles, period 2", "s_dihedral_angle_2_deg", "TORSION ANGLES, PERIOD 2    (DEGREES)"),
                            ("r.m.s.d.", "Torsion angles, period 3", "s_dihedral_angle_3_deg", "TORSION ANGLES, PERIOD 3    (DEGREES)"),
                            ("r.m.s.d.", "Torsion angles, period 6", "s_dihedral_angle_6_deg", "TORSION ANGLES, PERIOD 6    (DEGREES)"),
                            ("r.m.s.d.", "Chiral centres", "s_chiral_restr",                   "CHIRAL-CENTER RESTRAINTS       (A**3)"),
                            ("r.m.s.d.", "Planar groups", "s_planes",                          "GENERAL PLANES REFINED ATOMS      (A)"),
                            ("r.m.s.d.", "VDW nonbonded", "s_nbd",                             ""),
                            ("r.m.s.d.", "VDW torsion", "s_nbtor",                             ""),
                            ("r.m.s.d.", "VDW hbond", "s_hbond_nbd",                           ""),
                            ("r.m.s.d.", "VDW metal", "s_metal_ion",                           ""),
                            ("r.m.s.d.", "VDW dummy", "s_dummy_nbd",                           ""),
                            ("r.m.s.d.", "VDW nonbonded, symmetry", "s_symmetry_nbd",          ""),
                            ("r.m.s.d.", "VDW torsion, symmetry", "s_symmetry_nbtor",          ""),
                            ("r.m.s.d.", "VDW hbond, symmetry", "s_symmetry_hbond_nbd",        ""),
                            ("r.m.s.d.", "VDW metal, symmetry", "s_symmetry_metal_ion",        ""),
                            ("r.m.s.d.", "VDW dummy, symmetry", "s_symmetry_dummy_nbd",        "")):
            if k in stats["geom"]["summary"] and n in stats["geom"]["summary"][k]:
                rr = gemmi.RefinementInfo.Restr(l)
                rr.dev_ideal = round(stats["geom"]["summary"][k].get(n), 4)
                rr.count = stats["geom"]["summary"]["N restraints"].get(n)
                rr.weight = round(stats["geom"]["summary"]["Mn(sigma)"].get(n), 4)
                restr_stats.append(rr)
                if pl:
                    raw_remarks.append(f"REMARK   3   {pl}:{rr.count:6d} ;{rr.dev_ideal:6.3f} ;{rr.weight:6.3f}")
        ri.restr_stats = restr_stats
        raw_remarks.append("REMARK   3")
    if ll is not None:
        ri.id = ll.refine_id()
        ri.mean_b = round(numpy.mean([cra.atom.b_iso for cra in st[0].all()]), 2)
        if ll.b_aniso is not None:
            ri.aniso_b = ll.b_aniso
        for k, kd, nd in (("Rwork", "r_work", 4), ("Rfree", "r_free", 4), ("R", "r_all", 4),
                          ("FSCaverage", "fsc_work", 4),
                          ("FSCaverage_half1", "fsc_work", 4), ("FSCaverage_half2", "fsc_free", 4)):
            if k in stats["data"]["summary"]:
                setattr(ri, kd, round(stats["data"]["summary"][k], nd))
        bins = []
        n_all = 0
        for b in stats["data"]["binned"]:
            bri = gemmi.BasicRefinementInfo()
            bri.resolution_high = round(b["d_min"], 3)
            bri.resolution_low = round(b["d_max"], 3)
            for k, kd, nd in (("Rwork", "r_work", 4), ("Rfree", "r_free", 4),
                              ("R1work", "r_work", 4), ("R1free", "r_free", 4),
                              ("R", "r_all", 4), ("R1", "r_all", 4),
                              ("CCI", "cc_intensity_work", 4), ("CCF", "cc_fo_fc_work", 4),
                              ("CCIwork", "cc_intensity_work", 4), ("CCIfree", "cc_intensity_free", 4),
                              ("CCFwork", "cc_fo_fc_work", 4), ("CCFfree", "cc_fo_fc_free", 4),
                              ("fsc_FC_full", "fsc_work", 4), ("fsc_model", "fsc_work", 4),
                              ("fsc_model_half1", "fsc_work", 4), ("fsc_model_half2", "fsc_free", 4),
                              ("n_work", "work_set_count", 0), ("n_free", "rfree_set_count", 0),
                              ("n_obs", "reflection_count", 0), ("ncoeffs", "reflection_count", 0)):
                if k in b: setattr(bri, kd, round(b[k], nd))
            if "n_all" in b and "n_obs" in b:
                bri.completeness = round(b["n_obs"] / b["n_all"] * 100, 2)
                n_all += b["n_all"]
            bins.append(bri)
        ri.rfree_set_count = max(-1, sum(b.rfree_set_count for b in bins))
        ri.work_set_count = max(-1, sum(b.work_set_count for b in bins))
        ri.reflection_count = max(-1, sum(b.reflection_count for b in bins))
        ri.resolution_high = round(min(b.resolution_high for b in bins), 3)
        ri.resolution_low = round(max(b.resolution_low for b in bins), 3)
        if ri.reflection_count > 0 and n_all > 0:
            ri.completeness = round(ri.reflection_count / n_all * 100, 2)
        ri.bins = bins
        if ri.rfree_set_count > 0:
            ri.cross_validation_method = "THROUGHOUT"
    st.meta.refinement = [ri]
    st.raw_remarks = raw_remarks
# update_meta()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import argparse
import gemmi
import numpy
import pandas
import scipy.optimize
from servalcat.utils import logger
from servalcat import utils
from servalcat import ext

def calculate_obliquity(gv, twin_op):
    """
    Beforehand, the following calculation must be done
    gv = gemmi.GruberVector(cell, sg.centring_type(), True)
    gv.niggli_reduce()
    """
    def get_axis(r):
        eigenvalues, eigenvectors = numpy.linalg.eig(r)
        idx = numpy.argmin(numpy.abs(eigenvalues - 1))
        return eigenvectors[:,idx].real
    reduced_cell = gv.get_cell()
    orth = numpy.array(reduced_cell.orth.mat.tolist())
    frac = numpy.array(reduced_cell.frac.mat.tolist())
    op = gv.change_of_basis.inverse().combine(twin_op.as_xyz()).combine(gv.change_of_basis)
    r = numpy.array(op.rot) / op.DEN
    u = get_axis(r)
    h = get_axis(numpy.linalg.inv(r.transpose()))
    tau = h.dot(frac)
    t = orth.dot(u)
    obl_deg = numpy.rad2deg(numpy.arccos(numpy.clip(t.dot(tau) / numpy.linalg.norm(t) / numpy.linalg.norm(tau), -1, 1)))
    return obl_deg
# calculate_obliquity()

def find_twin_domains_from_data(hkldata, max_oblique=5, min_cc=0.2):
    logger.writeln("Finding possible twin operators from data")
    ops = gemmi.find_twin_laws(hkldata.cell, hkldata.sg, max_oblique, False)
    logger.writeln(f" {len(ops)} possible twin operator(s) found")
    #for op in ops:
    #    logger.writeln(f"  {op.triplet()}")
    if not ops:
        logger.writeln("")
        return None, None
    gv = gemmi.GruberVector(hkldata.cell, hkldata.sg.centring_type(), True)
    gv.niggli_reduce()
    twin_data = ext.TwinData()
    twin_data.setup(hkldata.miller_array(), hkldata.df.bin_ml, hkldata.sg, hkldata.cell, ops)
    if "I" in hkldata.df:
        Io = hkldata.df.I.to_numpy()
    else:
        Io = hkldata.df.FP.to_numpy()**2
    ccs, nums = [], []
    tmp = []
    for i_bin, bin_idxes in hkldata.binned("ml"):
        ccs.append([])
        nums.append([])
        rs = []
        for i_op, op in enumerate(ops):
            cc = r = numpy.nan
            ii = numpy.array(twin_data.pairs(i_op, i_bin))
            val = numpy.all(numpy.isfinite(Io[ii]), axis=1) if ii.size != 0 else []
            if numpy.sum(val) != 0:
                cc = numpy.corrcoef(Io[ii][val].T)[0,1]
                r = numpy.sum(numpy.abs(Io[ii][val, 0] - Io[ii][val, 1])) / numpy.sum(Io[ii][val])
            ccs[-1].append(cc)
            rs.append(r)
            nums[-1].append(len(val))
        tmp.append(rs + ccs[-1] + nums[-1])
    df = pandas.DataFrame(tmp, columns=[f"{n}_op{i+1}" for n in ("R", "CC", "num") for i in range(len(ops))])
    with logger.with_prefix(" "):
        logger.writeln(df.to_string(float_format="%.4f"))
    ccs = numpy.array(ccs)
    nums = numpy.array(nums)
    tmp = [{"Operator": "h,k,l",
            "Obliquity": 0,
            "R_twin_obs": 0,
            "CC_mean": 1}]
    for i_op, op in enumerate(ops):
        ii = numpy.array(twin_data.pairs(i_op))
        val = numpy.all(numpy.isfinite(Io[ii]), axis=1)
        if numpy.sum(val) == 0:
            r_obs = numpy.nan
        else:
            r_obs = numpy.sum(numpy.abs(Io[ii][val, 0] - Io[ii][val, 1])) / numpy.sum(Io[ii][val])
        good = numpy.isfinite(ccs[:,i_op])
        cc = numpy.sum(nums[good,i_op] * ccs[good,i_op]) / numpy.sum(nums[good,i_op])
        tmp.append({"Operator": op.as_hkl().triplet(),
                    "Obliquity": calculate_obliquity(gv, op),
                    "CC_mean": cc, 
                    "R_twin_obs": r_obs,
                    })
    df = pandas.DataFrame(tmp)
    with logger.with_prefix(" "):
        logger.writeln(df.to_string(float_format="%.2f"))

    sel = df["CC_mean"].to_numpy() > min_cc
    if sel[1:].sum() == 0:
        logger.writeln(" No possible twinning detected\n")
        return None, None
    
    if 0:#not sel.all():
        ops = [ops[i] for i in range(len(ops)) if sel[i+1]]
        logger.writeln(f"\n Twin operators after filtering small correlations (<= {min_cc})")
        df = df[sel]
        with logger.with_prefix(" "):
            logger.writeln(df.to_string(float_format="%.2f"))
        twin_data = ext.TwinData()
        twin_data.setup(hkldata.miller_array(), hkldata.df.bin_ml, hkldata.sg, hkldata.cell, ops)
    twin_data.alphas = [1. / len(twin_data.alphas) for _ in range(len(twin_data.alphas)) ]
    if "I" not in hkldata.df:
        logger.writeln('Generating "observed" intensities for twin refinement: Io = Fo**2, SigIo = 2*F*SigFo')
        hkldata.df["I"] = hkldata.df.FP**2
        hkldata.df["SIGI"] = 2 * hkldata.df.FP * hkldata.df.SIGFP
    logger.writeln("")
    return twin_data, df

# find_twin_domains_from_data()

def estimate_twin_fractions_from_model(twin_data, hkldata, min_alpha=0.02):
    logger.writeln("Estimating twin fractions")
    Ic = numpy.abs(twin_data.f_calc.sum(axis=1))**2
    idx_all = twin_data.twin_related(hkldata.sg)
    Ic_all = Ic[idx_all]
    Ic_all[(idx_all < 0).any(axis=1)] = numpy.nan
    rr = twin_data.obs_related_asu()
    tmp = []
    P_list, cc_oc_list, weight_list = [], [], []
    n_ops = len(twin_data.ops) + 1
    tidxes = numpy.triu_indices(n_ops, 1)
    if "CC*" in hkldata.binned_df["ml"]:
        logger.writeln(" data-correlations are corrected using CC*")
    for i_bin, bin_idxes in hkldata.binned("ml"): # XXX
        i_tmp = Ic_all[numpy.asarray(twin_data.bin)==i_bin,:]
        i_tmp = i_tmp[numpy.isfinite(i_tmp).all(axis=1)]
        P = numpy.corrcoef(i_tmp.T)
        iobs = hkldata.df.I.to_numpy()[bin_idxes]
        ic_bin = Ic[rr[bin_idxes,:]]
        val = numpy.isfinite(iobs) & numpy.isfinite(ic_bin).all(axis=1) & numpy.all(rr[bin_idxes,:]>=0, axis=1)
        iobs, ic_bin = iobs[val], ic_bin[val,:]
        cc_star = hkldata.binned_df["ml"]["CC*"][i_bin] if "CC*" in hkldata.binned_df["ml"] else 1
        if cc_star < 0.5:
            break
        cc_oc = [numpy.corrcoef(iobs, ic_bin[:,i])[0,1] / cc_star for i in range(n_ops)]
        P_list.append(P)
        cc_oc_list.append(cc_oc)
        weight_list.append(numpy.sum(val))
        frac_est = numpy.dot(numpy.linalg.pinv(P), cc_oc)
        frac_est /= frac_est.sum()
        tmp.append(P[tidxes].tolist() + cc_oc + [weight_list[-1]] + frac_est.tolist())

    good = numpy.logical_and(numpy.isfinite(P_list).any(axis=(1,2)), numpy.isfinite(cc_oc_list).any(axis=1))
    P_list = numpy.array(P_list)[good]
    cc_oc_list = numpy.array(cc_oc_list)[good]
    weight_list = numpy.array(weight_list)[good]
    P = numpy.average(P_list, axis=0, weights=weight_list)
    cc_oc = numpy.average(cc_oc_list, axis=0, weights=weight_list)
    frac_est = numpy.dot(numpy.linalg.pinv(P), cc_oc)
    frac_est = numpy.maximum(0, frac_est)
    frac_est /= frac_est.sum()
    df = pandas.DataFrame(tmp, columns=[f"cc_{i+1}_{j+1}" for i, j in zip(*tidxes)] +
                          [f"cc_o_{i+1}" for i in range(n_ops)] +
                          ["nref"] + [f"raw_est_{i+1}" for i in range(n_ops)])
    with logger.with_prefix(" "):
        logger.writeln(df.to_string(float_format="%.4f"))
    logger.write(" Final twin fraction estimate: ")
    logger.writeln(" ".join("%.4f"%x for x in frac_est))
    twin_data.alphas = frac_est

    if numpy.logical_and(0 < frac_est, frac_est < min_alpha).any():
        frac_est[frac_est < min_alpha] = 0.
        frac_est /= frac_est.sum()
        logger.write(" Small fraction removed: ")
        logger.writeln(" ".join("%.4f"%x for x in frac_est))
        twin_data.alphas = frac_est

    return df

def mlopt_twin_fractions(hkldata, twin_data, b_aniso):
    k_ani2_inv = 1 / hkldata.debye_waller_factors(b_cart=b_aniso)**2
    Io = hkldata.df.I.to_numpy(copy=True) * k_ani2_inv
    sigIo = hkldata.df.SIGI.to_numpy(copy=True) * k_ani2_inv
    def fun(x):
        twin_data.alphas = x
        twin_data.est_f_true(Io, sigIo, 10)
        ret = twin_data.ll(Io, sigIo)
        return ret
    def grad(x):
        twin_data.alphas = x
        twin_data.est_f_true(Io, sigIo, 10)
        return twin_data.ll_der_alpha(Io, sigIo, True)
    if 0:
        bak = [_ for _ in twin_data.alphas]
        with open("alpha_ll.csv", "w") as ofs:
            ofs.write("a,ll,ll_new,der1,der2,der_new1,der_new2\n")
            for a in numpy.linspace(0., 1.0, 100):
                x = [a, 1-a]
                twin_data.alphas = x
                twin_data.est_f_true(Io, sigIo, 100)
                f_new = twin_data.ll(Io, sigIo)
                f = twin_data.ll_rice()
                der = twin_data.ll_der_alpha(Io, sigIo, False)
                #der = [x - der[-1] for x in der[:-1]]
                der_new = twin_data.ll_der_alpha(Io, sigIo, True)
                #der_new = [x - der_new[-1] for x in der_new[:-1]]
                ofs.write(f"{a},{f},{f_new},{der[0]},{der[1]},{der_new[0]},{der_new[1]}\n")
            ofs.write("\n")
        twin_data.alphas = bak
    if 0:
        x0 = [x for x in twin_data.alphas]
        f0 = fun(x0)
        ader = grad(x0)
        
        print(f"{ader=}")
        for e in (1e-2, 1e-3, 1e-4, 1e-5):
            nder = []
            for i in range(len(x0)):
                x = [_ for _ in x0]
                x[i] += e
                f1 = fun(x)
                nder.append((f1 - f0) / e)
            print(f"{e=} {nder=}")
    
    logger.writeln("ML twin fraction refinement..")
    num_params = len(twin_data.alphas)
    A = numpy.ones((1, num_params))
    linear_constraint = scipy.optimize.LinearConstraint(A, [1.0], [1.0])
    bounds = scipy.optimize.Bounds(numpy.zeros(num_params), numpy.ones(num_params))
    logger.writeln(" starting with " + " ".join("%.4f"%x for x in twin_data.alphas))
    logger.writeln(f" f0= {fun(twin_data.alphas)}")
    res = scipy.optimize.minimize(fun=fun, x0=twin_data.alphas,
                                  bounds=bounds,
                                  constraints=[linear_constraint],
                                  jac=grad,
                                  #callback=lambda *x: logger.writeln(f"callback {x}"),
                                  )
    logger.writeln(" finished in {} iterations ({} evaluations)".format(res.nit, res.nfev))
    logger.writeln(f" f = {res.fun}")
    # ensure constraints
    alphas = numpy.clip(res.x, 0, 1)
    twin_data.alphas = list(alphas / alphas.sum())
    logger.write(" ML twin fraction estimate: ")
    logger.writeln(" ".join("%.4f"%x for x in twin_data.alphas))
